@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     flask_security.decorators
     ~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -6,12 +5,14 @@
     Flask-Security decorators module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
 from collections import namedtuple
+import datetime
 from functools import wraps
+import typing as t
 
 from flask import Response, _request_ctx_stack, abort, current_app, g, redirect, request
 from flask_login import current_user, login_required  # noqa: F401
@@ -20,17 +21,27 @@ from flask_wtf.csrf import CSRFError
 from werkzeug.local import LocalProxy
 from werkzeug.routing import BuildError
 
-from . import utils
+from .proxies import _security, DecoratedView
+from .utils import (
+    FsPermNeed,
+    config_value,
+    do_flash,
+    find_user,
+    get_message,
+    get_url,
+    check_and_update_authn_fresh,
+    json_error_response,
+    set_request_attr,
+    url_for_security,
+)
 
 # Convenient references
-_security = LocalProxy(lambda: current_app.extensions["security"])
-
 _csrf = LocalProxy(lambda: current_app.extensions["csrf"])
 
 BasicAuth = namedtuple("BasicAuth", "username, password")
 
 # NOTE: this is here for backwards compatibility, it is deprecated and
-# to be removed in 4.0
+# could be removed any time!
 _default_unauthenticated_html = """
     <h1>Unauthorized</h1>
     <p>The server could not verify that you are authorized to access the URL
@@ -52,39 +63,68 @@ def _get_unauthorized_response(text=None, headers=None):  # pragma: no cover
 
 
 def default_unauthn_handler(mechanisms, headers=None):
-    """ Default callback for failures to authenticate
+    """Default callback for failures to authenticate
 
-    If caller wants JSON - return 401
+    If caller wants JSON - return 401.
+    If caller wants BasicAuth - return 401 (the WWW-Authenticate header is set).
     Otherwise - assume caller is html and redirect if possible to a login view.
     We let Flask-Login handle this.
 
     """
-    unauthn_message, _ = utils.get_message("UNAUTHENTICATED")
+    headers = headers or {}
+    msg = get_message("UNAUTHENTICATED")[0]
 
-    if utils.config_value("BACKWARDS_COMPAT_UNAUTHN"):
+    if config_value("BACKWARDS_COMPAT_UNAUTHN"):
         return _get_unauthenticated_response(headers=headers)
     if _security._want_json(request):
-        # TODO can/should we response with a WWW-Authenticate Header in all cases?
-        payload = utils.json_error_response(errors=unauthn_message)
+        payload = json_error_response(errors=msg)
         return _security._render_json(payload, 401, headers, None)
+
+    # Basic-Auth is often used to provide a browser based login form and then the
+    # browser will always add the BasicAuth credentials. For that to work we need to
+    # return 401 and not redirect to our login view.
+    if "WWW-Authenticate" in headers:
+        return Response(msg, 401, headers)
     return _security.login_manager.unauthorized()
 
 
-def default_unauthz_handler(func, params):
-    unauthz_message, unauthz_message_type = utils.get_message("UNAUTHORIZED")
+def default_reauthn_handler(within, grace):
+    """Default callback for 'freshness' related authn failures.
+
+    If caller wants JSON - return 401
+    Otherwise - assume caller is html and redirect if possible to configured view.
+
+    """
+    m, c = get_message("REAUTHENTICATION_REQUIRED")
+
     if _security._want_json(request):
-        payload = utils.json_error_response(errors=unauthz_message)
+        is_us = config_value("UNIFIED_SIGNIN")
+        payload = json_error_response(errors=m)
+        payload["reauth_required"] = True
+        payload["unified_signin_enabled"] = is_us
+        return _security._render_json(payload, 401, None, None)
+
+    view = "us_verify" if config_value("UNIFIED_SIGNIN") else "verify"
+    do_flash(m, c)
+    redirect_url = url_for_security(view, next=request.url)
+    return redirect(redirect_url)
+
+
+def default_unauthz_handler(func, params):
+    unauthz_message, unauthz_message_type = get_message("UNAUTHORIZED")
+    if _security._want_json(request):
+        payload = json_error_response(errors=unauthz_message)
         return _security._render_json(payload, 403, None, None)
-    view = utils.config_value("UNAUTHORIZED_VIEW")
+    view = config_value("UNAUTHORIZED_VIEW")
     if view:
         if callable(view):
             view = view()
         else:
             try:
-                view = utils.get_url(view)
+                view = get_url(view)
             except BuildError:
                 view = None
-        utils.do_flash(unauthz_message, unauthz_message_type)
+        do_flash(unauthz_message, unauthz_message_type)
         redirect_to = "/"
         if request.referrer and not request.referrer.split("?")[0].endswith(
             request.path
@@ -98,7 +138,7 @@ def default_unauthz_handler(func, params):
 def _check_token():
     # N.B. this isn't great Flask-Login 0.5.0 made this protected
     # Issue https://github.com/maxcountryman/flask-login/issues/471
-    # Was files to restore public access. We want to call this via
+    # was filed to restore public access. We want to call this via
     # login_manager in case someone has overridden the login_manager which we
     # allow.
     if hasattr(_security.login_manager, "request_callback"):
@@ -110,7 +150,7 @@ def _check_token():
     if user and user.is_authenticated:
         app = current_app._get_current_object()
         _request_ctx_stack.top.user = user
-        identity_changed.send(app, identity=Identity(user.id))
+        identity_changed.send(app, identity=Identity(user.fs_uniquifier))
         return True
 
     return False
@@ -120,20 +160,22 @@ def _check_http_auth():
     auth = request.authorization or BasicAuth(username=None, password=None)
     if not auth.username:
         return False
-    user = _security.datastore.get_user(auth.username)
+    user = find_user(auth.username)
+    if user and not user.active:
+        return False
 
     if user and user.verify_and_update_password(auth.password):
         _security.datastore.commit()
         app = current_app._get_current_object()
         _request_ctx_stack.top.user = user
-        identity_changed.send(app, identity=Identity(user.id))
+        identity_changed.send(app, identity=Identity(user.fs_uniquifier))
         return True
 
     return False
 
 
-def handle_csrf(method):
-    """ Invoke CSRF protection based on authentication method.
+def handle_csrf(method: t.Optional[str]) -> None:
+    """Invoke CSRF protection based on authentication method.
 
     Usually this is called as part of a decorator, but if that isn't
     appropriate, endpoint code can call this directly.
@@ -162,18 +204,21 @@ def handle_csrf(method):
     ):
         return
 
-    if utils.config_value("CSRF_PROTECT_MECHANISMS"):
-        if method in utils.config_value("CSRF_PROTECT_MECHANISMS"):
-            _csrf.protect()
+    if config_value("CSRF_PROTECT_MECHANISMS"):
+        if method in config_value("CSRF_PROTECT_MECHANISMS"):
+            _csrf.protect()  # type: ignore
         else:
             _request_ctx_stack.top.fs_ignore_csrf = True
 
 
-def http_auth_required(realm):
+def http_auth_required(realm: t.Any) -> DecoratedView:
     """Decorator that protects endpoints using Basic HTTP authentication.
 
     :param realm: optional realm name
 
+    If authentication fails, then a 401 with the 'WWW-Authenticate' header set will be
+    returned.
+
     Once authenticated, if so configured, CSRF protection will be tested.
     """
 
@@ -182,6 +227,7 @@ def http_auth_required(realm):
         def wrapper(*args, **kwargs):
             if _check_http_auth():
                 handle_csrf("basic")
+                set_request_attr("fs_authn_via", "basic")
                 return fn(*args, **kwargs)
             if _security._unauthorized_callback:
                 return _security._unauthorized_callback()
@@ -197,7 +243,7 @@ def http_auth_required(realm):
     return decorator
 
 
-def auth_token_required(fn):
+def auth_token_required(fn: DecoratedView) -> DecoratedView:
     """Decorator that protects endpoints using token authentication. The token
     should be added to the request by the client by using a query string
     variable with a name equal to the configuration value of
@@ -211,16 +257,21 @@ def auth_token_required(fn):
     def decorated(*args, **kwargs):
         if _check_token():
             handle_csrf("token")
+            set_request_attr("fs_authn_via", "token")
             return fn(*args, **kwargs)
         if _security._unauthorized_callback:
             return _security._unauthorized_callback()
         else:
             return _security._unauthn_handler(["token"])
 
-    return decorated
+    return t.cast(DecoratedView, decorated)
 
 
-def auth_required(*auth_methods):
+def auth_required(
+    *auth_methods: t.Union[str, t.Callable[[], t.List[str]], None],
+    within: t.Union[int, float, t.Callable[[], datetime.timedelta]] = -1,
+    grace: t.Optional[t.Union[int, float, t.Callable[[], datetime.timedelta]]] = None
+) -> DecoratedView:
     """
     Decorator that protects endpoints through multiple mechanisms
     Example::
@@ -231,7 +282,26 @@ def auth_required(*auth_methods):
             return 'Dashboard'
 
     :param auth_methods: Specified mechanisms (token, basic, session). If not specified
-        then all current available mechanisms will be tried.
+        then all current available mechanisms (except "basic") will be tried. A callable
+        can also be passed (useful if you need app/request context). The callable
+        must return a list.
+    :param within: Add 'freshness' check to authentication. Is either an int
+        specifying # of minutes, or a callable that returns a timedelta. For timedeltas,
+        timedelta.total_seconds() is used for the calculations:
+
+            - If > 0, then the caller must have authenticated within the time specified
+              (as measured using the session cookie).
+            - If 0 and not within the grace period (see below) the caller will
+              always be redirected to re-authenticate.
+            - If < 0 (the default) no freshness check is performed.
+
+        Note that Basic Auth, by definition, is always 'fresh' and will never result in
+        a redirect/error.
+    :param grace: Add a grace period for freshness checks. As above, either an int
+        or a callable returning a timedelta. If not specified then
+        :py:data:`SECURITY_FRESHNESS_GRACE_PERIOD` is used. The grace period allows
+        callers to complete the required operations w/o being prompted again.
+        See :meth:`flask_security.check_and_update_authn_fresh` for details.
 
     Note that regardless of order specified - they will be tried in the following
     order: token, session, basic.
@@ -239,52 +309,98 @@ def auth_required(*auth_methods):
     The first mechanism that succeeds is used, following that, depending on
     configuration, CSRF protection will be tested.
 
-    On authentication failure :meth:`.Security.unauthorized_callback` (deprecated)
+    On authentication failure `.Security.unauthorized_callback` (deprecated)
     or :meth:`.Security.unauthn_handler` will be called.
 
+    As a side effect, upon successful authentication, the request global
+     ``fs_authn_via`` will be set to the method ("basic", "token", "session")
+
+    .. note::
+        If "basic" is specified in addition to other methods, then if authentication
+        fails, a 401 with the "WWW-Authenticate" header will be returned - rather than
+        being redirected to the login view.
+
     .. versionchanged:: 3.3.0
        If ``auth_methods`` isn't specified, then all will be tried. Authentication
        mechanisms will always be tried in order of ``token``, ``session``, ``basic``
        regardless of how they are specified in the ``auth_methods`` parameter.
+
+    .. versionchanged:: 3.4.0
+        Added ``within`` and ``grace`` parameters to enforce a freshness check.
+
+    .. versionchanged:: 3.4.4
+        If ``auth_methods`` isn't specified try all mechanisms EXCEPT ``basic``.
+
+    .. versionchanged:: 4.0.0
+        auth_methods can be passed as a callable.
+
     """
+
     login_mechanisms = {
         "token": lambda: _check_token(),
         "session": lambda: current_user.is_authenticated,
         "basic": lambda: _check_http_auth(),
     }
     mechanisms_order = ["token", "session", "basic"]
-    if not auth_methods:
-        auth_methods = {"basic", "session", "token"}
-    else:
-        auth_methods = [am for am in auth_methods]
 
     def wrapper(fn):
         @wraps(fn)
-        def decorated_view(*args, **kwargs):
+        def decorated_view(
+            *args, auth_methods=auth_methods, within=within, grace=grace, **kwargs
+        ):
+            if callable(within):
+                within = within()
+            else:
+                within = datetime.timedelta(minutes=within)
+            if grace is None:
+                grace = config_value("FRESHNESS_GRACE_PERIOD")
+            elif callable(grace):
+                grace = grace()
+            else:
+                grace = datetime.timedelta(minutes=grace)
+
+            if not auth_methods:
+                ams = {"session", "token"}
+            else:
+                ams = []
+                for am in auth_methods:
+                    if callable(am):
+                        ams.extend(am())
+                    else:
+                        ams.append(am)
+
             h = {}
+            if "basic" in ams:
+                r = _security.default_http_auth_realm
+                h["WWW-Authenticate"] = 'Basic realm="%s"' % r
             mechanisms = [
                 (method, login_mechanisms.get(method))
                 for method in mechanisms_order
-                if method in auth_methods
+                if method in ams
             ]
             for method, mechanism in mechanisms:
                 if mechanism and mechanism():
+                    # successfully authenticated. Basic auth is by definition 'fresh'.
+                    # Note that using token auth is ok - but caller still has to pass
+                    # in a session cookie...
+                    if not check_and_update_authn_fresh(within, grace, method):
+                        return _security._reauthn_handler(within, grace)
                     handle_csrf(method)
+                    set_request_attr("fs_authn_via", method)
                     return fn(*args, **kwargs)
-                elif method == "basic":
-                    r = _security.default_http_auth_realm
-                    h["WWW-Authenticate"] = 'Basic realm="%s"' % r
             if _security._unauthorized_callback:
                 return _security._unauthorized_callback()
             else:
-                return _security._unauthn_handler(auth_methods, headers=h)
+                return _security._unauthn_handler(ams, headers=h)
 
         return decorated_view
 
     return wrapper
 
 
-def unauth_csrf(fall_through=False):
+def unauth_csrf(
+    fall_through: bool = False,
+) -> DecoratedView:
     """Decorator for endpoints that don't need authentication
     but do want CSRF checks (available via Header rather than just form).
     This is required when setting *WTF_CSRF_CHECK_DEFAULT* = **False** since in that
@@ -320,7 +436,7 @@ def unauth_csrf(fall_through=False):
                 return fn(*args, **kwargs)
 
             if (
-                utils.config_value("CSRF_IGNORE_UNAUTH_ENDPOINTS")
+                config_value("CSRF_IGNORE_UNAUTH_ENDPOINTS")
                 and not current_user.is_authenticated
             ):
                 _request_ctx_stack.top.fs_ignore_csrf = True
@@ -338,7 +454,7 @@ def unauth_csrf(fall_through=False):
     return wrapper
 
 
-def roles_required(*roles):
+def roles_required(*roles: str) -> DecoratedView:
     """Decorator which specifies that a user must have all the specified roles.
     Example::
 
@@ -370,7 +486,7 @@ def roles_required(*roles):
     return wrapper
 
 
-def roles_accepted(*roles):
+def roles_accepted(*roles: str) -> DecoratedView:
     """Decorator which specifies that a user must have at least one of the
     specified roles. Example::
 
@@ -388,7 +504,7 @@ def roles_accepted(*roles):
     def wrapper(fn):
         @wraps(fn)
         def decorated_view(*args, **kwargs):
-            perm = Permission(*[RoleNeed(role) for role in roles])
+            perm = Permission(*(RoleNeed(role) for role in roles))
             if perm.can():
                 return fn(*args, **kwargs)
             if _security._unauthorized_callback:
@@ -401,7 +517,7 @@ def roles_accepted(*roles):
     return wrapper
 
 
-def permissions_required(*fsperms):
+def permissions_required(*fsperms: str) -> DecoratedView:
     """Decorator which specifies that a user must have all the specified permissions.
     Example::
 
@@ -423,7 +539,7 @@ def permissions_required(*fsperms):
     def wrapper(fn):
         @wraps(fn)
         def decorated_view(*args, **kwargs):
-            perms = [Permission(utils.FsPermNeed(fsperm)) for fsperm in fsperms]
+            perms = [Permission(FsPermNeed(fsperm)) for fsperm in fsperms]
             for perm in perms:
                 if not perm.can():
                     if _security._unauthorized_callback:
@@ -440,7 +556,7 @@ def permissions_required(*fsperms):
     return wrapper
 
 
-def permissions_accepted(*fsperms):
+def permissions_accepted(*fsperms: str) -> DecoratedView:
     """Decorator which specifies that a user must have at least one of the
     specified permissions. Example::
 
@@ -462,7 +578,7 @@ def permissions_accepted(*fsperms):
     def wrapper(fn):
         @wraps(fn)
         def decorated_view(*args, **kwargs):
-            perm = Permission(*[utils.FsPermNeed(fsperm) for fsperm in fsperms])
+            perm = Permission(*(FsPermNeed(fsperm) for fsperm in fsperms))
             if perm.can():
                 return fn(*args, **kwargs)
             if _security._unauthorized_callback:
@@ -475,7 +591,7 @@ def permissions_accepted(*fsperms):
     return wrapper
 
 
-def anonymous_user_required(f):
+def anonymous_user_required(f: DecoratedView) -> DecoratedView:
     """Decorator which requires that caller NOT be logged in.
     If a logged in user accesses an endpoint protected with this decorator
     they will be redirected to the *SECURITY_POST_LOGIN_VIEW*.
@@ -489,12 +605,12 @@ def anonymous_user_required(f):
     def wrapper(*args, **kwargs):
         if current_user.is_authenticated:
             if _security._want_json(request):
-                payload = utils.json_error_response(
-                    errors=utils.get_message("ANONYMOUS_USER_REQUIRED")[0]
+                payload = json_error_response(
+                    errors=get_message("ANONYMOUS_USER_REQUIRED")[0]
                 )
                 return _security._render_json(payload, 400, None, None)
             else:
-                return redirect(utils.get_url(_security.post_login_view))
+                return redirect(get_url(_security.post_login_view))
         return f(*args, **kwargs)
 
-    return wrapper
+    return t.cast(DecoratedView, wrapper)
